import * as G from "geojson";
import { LineTransectFeature, LineTransectGeometry } from "./line-transect.defs";
export declare type CRSString = "WGS84" | "EPSG:2393" | "EPSG:3067";
export declare type GeometryFormat = "GeoJSON" | "WKT" | "ISO 6709";
export interface GeoJSONValidationErrors {
    [path: string]: {
        message: string;
    };
}
export declare function reverseCoordinate(c: [number, number]): [number, number];
export declare function convertLatLng(latlng: [number, number], from: string, to: string, validate?: boolean): any;
export declare function updateImmutablyRecursivelyWith(obj: any, fn: (key: string, value: any) => any): any;
export declare function convertGeoJSON(geoJSON: G.GeoJSON, from: string, to: string, validate?: boolean | "errors"): G.GeoJSON;
export declare function circleToPolygon(center: any, radius: any, numberOfSegments: any): {
    type: string;
    coordinates: any[][];
};
export declare function standardizeGeoJSON(geoJSON: G.GeoJSON): G.GeoJSON;
export declare function geoJSONToTextualFormatWith(geoJSON: G.GeoJSON, name: string, latLngCoordConverter: (latlng: number[]) => string, coordinateJoiner: (latlng: number[]) => string, coordinateStrToPoint: (coordinate: string) => string, coordinateStrToLine: (coordinate: string) => string, coordinateStrToPolygon: (coordinate: string) => string): string;
export declare function geoJSONToISO6709(geoJSON: G.GeoJSON): string;
export declare function ISO6709ToGeoJSON(ISO6709: string): G.FeatureCollection;
export declare function geoJSONToWKT(geoJSON: G.GeoJSON): string;
export declare function WKTToGeoJSON(WKT: string): G.FeatureCollection;
export declare function latLngOrTupleToTuple(latLng: any): any;
export declare function latLngTuplesEqual(first: any, second: any): boolean;
export declare function latLngTuplesDistance(first: any, second: any): number;
declare function latLngSegmentsToGeoJSONGeometry(_lines: any): LineTransectGeometry;
export { latLngSegmentsToGeoJSONGeometry };
export declare function geoJSONLineToLatLngSegmentArrays(geometry: any): any;
export declare function detectFormat(data: any): GeometryFormat;
export declare function detectCRSFromLatLng(latLng: any, allowGrid?: boolean): CRSString;
export declare function detectCRS(data: string | G.GeoJSON, allowGrid?: boolean): string;
export declare function convertAnyToWGS84GeoJSON(data: string | G.GeoJSON, validate?: boolean | "errors"): G.GeoJSON;
/**
 * Geometric data conversion between supported formats & coordinate systems.
 * @param input Geometric data in GeoJSON, WKT or ISO 6709.
 * @param outputFormat The output format - one of "GeoJSON", "WKT" or "ISO 6709".
 * @param outputCRS The output coordinate system. Can be any system that proj4 supports, or EPSG:2393 or EPSG:3067.
 * @param validate If true, validates strictly with warning level errors. If "errors", doesn't validate with warning level errors. If false, doesn't validate at all.
 */
export declare function convert(input: string | G.GeoJSON, outputFormat: "WKT" | "ISO 6709", outputCRS: string, validate?: boolean | "errors"): string;
export declare function convert(input: string | G.GeoJSON, outputFormat: "GeoJSON", outputCRS: string, validate?: boolean | "errors"): G.FeatureCollection;
export declare function convert(input: string | G.GeoJSON, outputFormat: GeometryFormat, outputCRS: string, validate?: boolean | "errors"): G.FeatureCollection;
export declare function convert(input: string | G.GeoJSON, outputFormat: GeometryFormat, outputCRS: string, validate?: boolean | "errors"): string;
export declare function validateGeoJSON(geoJSON: string | G.GeoJSON, crs?: string, warnings?: boolean): {
    errors: Error[];
    geoJSON: G.GeoJSON;
};
export declare function coordinatesAreClockWise(coordinates: G.Position[]): boolean;
export declare function getCRSObjectForGeoJSON(geoJSON: G.GeoJSON, crs: string): {
    type: "name";
    properties: {
        name: string;
    };
};
export declare class LajiMapError extends Error {
    _lajiMapError: boolean;
    translationKey: string;
    lineIdx?: number;
    constructor(message: string, translationKey: string, lineIdx?: number);
    stringify: (translations: any) => string;
}
export declare class LajiMapGeoJSONConversionError extends Error {
    _lajiMapGeoJSONConversionError: boolean;
    errors: Error[];
    geoJSON: G.GeoJSON;
    constructor(message: string, errors: Error[], geoJSON: G.GeoJSON);
}
export declare class LajiMapGeoJSONError extends LajiMapError {
    translationKey: string;
    path: string;
    fixable: boolean;
    constructor(translationKey: string, path?: string, fixable?: boolean);
    render: (translations: any) => HTMLLIElement;
}
export declare type HalfString = "lat" | "lng";
export declare class CoordinateError extends LajiMapGeoJSONError {
    latOrLng: string;
    half: HalfString;
    constructor(translationKey: string, latOrLng: string, half?: HalfString, path?: string);
    stringify: (translations: any) => string;
    render: (translations: any) => HTMLLIElement;
}
export declare function stringifyLajiMapError(error: LajiMapError, translations: any): string;
export declare function renderLajiMapError(error: LajiMapError, translations: any, elemType?: string): HTMLElement;
export declare function parseJSON(json: string): any;
interface CoordinateValidator {
    regexp: RegExp;
    range: number[];
    formatter(value: string): string;
}
declare const wgs84Validator: CoordinateValidator[];
declare const ykjValidator: CoordinateValidator[];
declare const etrsTm35FinValidator: CoordinateValidator[];
declare const etrsValidator: CoordinateValidator[];
declare const ykjGridStrictValidator: CoordinateValidator[];
declare const ykjGridValidator: CoordinateValidator[];
declare const etrsTm35FinGridStrictValidator: {
    regexp: RegExp;
    formatter: (value: any) => any;
    range: number[];
}[];
export { wgs84Validator, ykjValidator, ykjGridValidator, ykjGridStrictValidator, etrsTm35FinValidator, etrsValidator, etrsTm35FinGridStrictValidator };
export declare function validateLatLng(latlng: string[], latLngValidator: CoordinateValidator[], throwError?: boolean): boolean;
export declare function roundMeters(meters: number, accuracy?: number): number;
export declare function createTextInput(): HTMLInputElement;
export declare function createTextArea(rows?: number, cols?: number): HTMLTextAreaElement;
export declare function isObject(obj: any): boolean;
export declare function combineColors(...colors: any[]): string;
export declare function getLineTransectStartEndDistancesForIdx(LTFeature: LineTransectFeature, idx: number, round?: number): number[];
export declare const capitalizeFirstLetter: (s: string) => string;
export declare function anyToFeatureCollection(_geoJSON: G.GeoJSON): G.FeatureCollection;
export declare function flattenMultiLineStringsAndMultiPolygons(features: G.Feature[]): G.Feature[];
export declare const geoJSONAsFeatureCollection: (geoJSON: G.GeoJSON) => G.FeatureCollection;
export declare function latLngGridToGeoJSON(latlngStr: [string, string]): G.Feature<G.Point | G.Polygon>;
