"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const test_utils_1 = require("./test-utils");
const utils_1 = require("laji-map/lib/utils");
const globals_1 = require("laji-map/lib/globals");
describe("Draw upload control", () => {
    let map, control;
    beforeAll(() => __awaiter(void 0, void 0, void 0, function* () {
        map = yield test_utils_1.createMap({
            draw: true,
            controls: {
                draw: {
                    upload: true
                }
            }
        });
        control = map.getCoordinateUploadControl();
    }));
    const getGeometry = () => map.e("getDraw().featureCollection.features[0].geometry");
    const asGeoJSON = (lat, lng) => JSON.stringify({ type: "Point", coordinates: [lng, lat] });
    const asISO6709 = (lat, lng) => `${lat}:${lng}/`;
    const asWKT = (lat, lng) => `POINT(${lng} ${lat})`;
    const pointTestsDescriptions = [
        { name: "GeoJSON", latLngToFormat: asGeoJSON },
        { name: "ISO 6709", latLngToFormat: asISO6709 },
        { name: "WKT", latLngToFormat: asWKT }
    ];
    it("opens on click", () => __awaiter(void 0, void 0, void 0, function* () {
        const $button = control.$getButton();
        yield $button.click();
        expect(yield control.$getContainer().isPresent()).toBe(true);
    }));
    it("is disabled when no input", () => __awaiter(void 0, void 0, void 0, function* () {
        const $submit = control.$getSubmit();
        expect(yield $submit.isEnabled()).toBe(false);
    }));
    it("is disabled with invalid latlng", () => __awaiter(void 0, void 0, void 0, function* () {
        const $submit = control.$getSubmit();
        yield control.type("sdfsdf");
        expect(yield $submit.isEnabled()).toBe(false);
    }));
    it("closes on close button click", () => __awaiter(void 0, void 0, void 0, function* () {
        yield control.$getCloseButton().click();
        expect(yield control.$getContainer().isPresent()).toBe(false);
    }));
    describe("accepts", () => {
        beforeEach(() => __awaiter(void 0, void 0, void 0, function* () {
            const $button = control.$getButton();
            yield $button.click();
        }));
        afterEach(() => __awaiter(void 0, void 0, void 0, function* () {
            if (yield control.$getContainer().isPresent()) {
                yield control.$getCloseButton().click();
            }
        }));
        for (const { name, latLngToFormat } of pointTestsDescriptions) {
            describe(name, () => {
                it("WGS84 point", () => __awaiter(void 0, void 0, void 0, function* () {
                    const [lat, lng] = [60, 25];
                    yield control.type(latLngToFormat(lat, lng));
                    expect(yield control.getCRS()).toBe("WGS84");
                    expect(yield control.getFormat()).toBe(name);
                    yield control.$getSubmit().click();
                    const mapCoordinates = (yield getGeometry()).coordinates;
                    expect(mapCoordinates[0]).toBe(lng);
                    expect(mapCoordinates[1]).toBe(lat);
                }));
                it("YKJ point", () => __awaiter(void 0, void 0, void 0, function* () {
                    const [lat, lng] = [6666666, 3333333];
                    const wgs84LatLng = test_utils_1.ykjToWgs84([lat, lng]);
                    yield control.type(latLngToFormat(lat, lng));
                    expect(yield control.getCRS()).toBe("YKJ");
                    expect(yield control.getFormat()).toBe(name);
                    yield control.$getSubmit().click();
                    const geometry = yield getGeometry();
                    expect(geometry.type).toBe("Point");
                    expect(utils_1.reverseCoordinate(geometry.coordinates)).toEqual(wgs84LatLng);
                }));
                it("ETRS point", () => __awaiter(void 0, void 0, void 0, function* () {
                    const [lat, lng] = [6666666, 333333];
                    const wgs84LatLng = test_utils_1.etrsToWgs84([lat, lng]);
                    yield control.type(latLngToFormat(lat, lng));
                    expect(yield control.getCRS()).toBe("ETRS-TM35FIN");
                    expect(yield control.getFormat()).toBe(name);
                    yield control.$getSubmit().click();
                    const geometry = yield getGeometry();
                    expect(geometry.type).toBe("Point");
                    expect(utils_1.reverseCoordinate(geometry.coordinates)).toEqual(wgs84LatLng);
                }));
            });
        }
        it("GeoJSON YKJ point when CRS marked", () => __awaiter(void 0, void 0, void 0, function* () {
            const [lat, lng] = [6666666, 3333333];
            const wgs84LatLng = test_utils_1.ykjToWgs84([lat, lng]);
            const geoJSON = JSON.parse(asGeoJSON(lat, lng));
            geoJSON.crs = {
                type: "name",
                properties: {
                    name: globals_1.EPSG2393String
                }
            };
            yield control.type(JSON.stringify(geoJSON));
            expect(yield control.getCRS()).toBe("YKJ");
            expect(yield control.getFormat()).toBe("GeoJSON");
            yield control.$getSubmit().click();
            const geometry = yield getGeometry();
            expect(geometry.type).toBe("Point");
            expect(utils_1.reverseCoordinate(geometry.coordinates)).toEqual(wgs84LatLng);
        }));
        it("ISO6709 point with CRS on same line", () => __awaiter(void 0, void 0, void 0, function* () {
            const [lat, lng] = [6666666, 333333];
            const wgs84LatLng = test_utils_1.etrsToWgs84([lat, lng]);
            yield control.type(`${asISO6709(lat, lng)}CRSEPSG:3067`);
            expect(yield control.getCRS()).toBe("ETRS-TM35FIN");
            expect(yield control.getFormat()).toBe("ISO 6709");
            yield control.$getSubmit().click();
            const geometry = yield getGeometry();
            expect(geometry.type).toBe("Point");
            expect(utils_1.reverseCoordinate(geometry.coordinates)).toEqual(wgs84LatLng);
        }));
        it("ISO6709 points with CRS on own line", () => __awaiter(void 0, void 0, void 0, function* () {
            const [lat, lng] = [6666666, 333333];
            const wgs84LatLng = test_utils_1.etrsToWgs84([lat, lng]);
            yield control.type(`${asISO6709(lat, lng)}\nCRSEPSG:3067`);
            expect(yield control.getCRS()).toBe("ETRS-TM35FIN");
            expect(yield control.getFormat()).toBe("ISO 6709");
            yield control.$getSubmit().click();
            const geometry = yield getGeometry();
            expect(geometry.type).toBe("Point");
            expect(utils_1.reverseCoordinate(geometry.coordinates)).toEqual(wgs84LatLng);
        }));
    });
});
