"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const test_utils_1 = require("./test-utils");
// Internal logic tested because the later tests rely heavily on it.
describe("Internal logic of zoom level normalization", () => {
    let map;
    it("uses the real zoom level for Finnish layer", () => __awaiter(void 0, void 0, void 0, function* () {
        map = yield test_utils_1.createMap({ tileLayerName: "taustakartta", zoom: 0 });
        yield expect(yield map.e("map.getZoom()")).toBe(0);
        yield expect(yield map.e("getNormalizedZoom()")).toBe(0);
    }));
    it("uses offset of 3 for world map layer", () => __awaiter(void 0, void 0, void 0, function* () {
        map = yield test_utils_1.createMap({ tileLayerName: "openStreetMap", zoom: 0 });
        yield expect(yield map.e("map.getZoom()")).toBe(3);
        yield expect(yield map.e("getNormalizedZoom()")).toBe(0);
    }));
    it("getNormalizedZoom() returns updated zoom after changing zoom", () => __awaiter(void 0, void 0, void 0, function* () {
        map = yield test_utils_1.createMap({ tileLayerName: "taustakartta", zoom: 0 });
        yield map.e("setNormalizedZoom(2)");
        yield expect(yield map.e("getNormalizedZoom()")).toBe(2);
    }));
    it("getNormalizedZoom() returns updated zoom after changing projection", () => __awaiter(void 0, void 0, void 0, function* () {
        map = yield test_utils_1.createMap({ tileLayerName: "taustakartta", zoom: 0 });
        yield map.e("setTileLayerByName(\"openStreetMap\")");
        yield expect(yield map.e("getNormalizedZoom()")).toBe(0);
    }));
});
describe("Initializing", () => {
    describe("default view", () => {
        let map = undefined;
        beforeAll(() => __awaiter(void 0, void 0, void 0, function* () {
            map = yield test_utils_1.createMap();
        }));
        it("opens with taustakartta", () => __awaiter(void 0, void 0, void 0, function* () {
            yield expect(yield map.e("tileLayerName")).toBe("taustakartta");
        }));
        it("opens with zoom level 2", () => __awaiter(void 0, void 0, void 0, function* () {
            yield expect(yield map.e("getNormalizedZoom()")).toBe(2);
        }));
    });
    describe("world map (openStreetMap)", () => {
        let map = undefined;
        beforeAll(() => __awaiter(void 0, void 0, void 0, function* () {
            map = yield test_utils_1.createMap({ tileLayerName: "openStreetMap" });
        }));
        it("opens with openStreetMap", () => __awaiter(void 0, void 0, void 0, function* () {
            yield expect(yield map.e("tileLayerName")).toBe("openStreetMap");
        }));
        it("opens with zoom level 2", () => __awaiter(void 0, void 0, void 0, function* () {
            yield expect(yield map.e("getNormalizedZoom()")).toBe(2);
        }));
    });
    it("with location outside Finland using Finnish tilelayer, zoom stays the same after swapping to world map view", () => __awaiter(void 0, void 0, void 0, function* () {
        const congo = {
            lat: 79.3499057749654,
            lng: 21.160612106323246
        };
        const map = yield test_utils_1.createMap({ tileLayerName: "taustakartta", zoom: 4, center: congo });
        yield expect(yield map.e("getNormalizedZoom()")).toBe(4);
    }));
    it("openStreetMap can have negative zoom level", () => __awaiter(void 0, void 0, void 0, function* () {
        const map = yield test_utils_1.createMap({ tileLayerName: "openStreetMap", zoom: -3 });
        yield expect(yield map.e("getNormalizedZoom()")).toBe(-3);
    }));
    it("taustakartta is zoomed to it's minimum (0) with negative value", () => __awaiter(void 0, void 0, void 0, function* () {
        const map = yield test_utils_1.createMap({ tileLayerName: "taustakartta", zoom: -3 });
        yield expect(yield map.e("getNormalizedZoom()")).toBe(0);
    }));
});
describe("Zooms to data", () => {
    const [lng, lat] = [25, 62];
    let zoom;
    it("when using openStreetMap", () => __awaiter(void 0, void 0, void 0, function* () {
        const map = yield test_utils_1.createMap({
            tileLayerName: "openStreetMap",
            zoomToData: true,
            data: { geoData: { type: "Point", coordinates: [lng, lat] } }
        });
        const center = yield map.e("map.getCenter()");
        zoom = yield map.e("getNormalizedZoom()");
        yield expect(center.lat).toBeCloseTo(lat);
        yield expect(center.lng).toBeCloseTo(lng);
    }));
    it("when using taustakartta", () => __awaiter(void 0, void 0, void 0, function* () {
        const map = yield test_utils_1.createMap({
            tileLayerName: "taustakartta",
            zoomToData: true,
            data: { geoData: { type: "Point", coordinates: [lng, lat] } }
        });
        const center = yield map.e("map.getCenter()");
        yield expect(center.lat).toBeCloseTo(lat);
        yield expect(center.lng).toBeCloseTo(lng);
        // Test that zoom is kept the same as with openStreetMap in prev test
        yield expect(yield map.e("getNormalizedZoom()")).toBeCloseTo(zoom);
    }));
    describe("correct when data, draw and lineTransect given", () => {
        let [north, east] = [63, 25];
        const [minNorth, minEast] = [north, east];
        const travel = new test_utils_1.LatLngTraveller(north, east, { onlyForward: true });
        const data = { geoData: { type: "Point", coordinates: [east, north] } };
        const draw = { geoData: { type: "Point", coordinates: travel.northEast(0.005, 0.005) } };
        const lineTransect = {
            feature: {
                type: "Feature",
                properties: {},
                geometry: {
                    type: "LineString",
                    coordinates: [
                        travel.northEast(0.005, 0.005),
                        travel.northEast(0.005, 0.005)
                    ]
                }
            }
        };
        const options = {
            zoomToData: true,
            draw,
            data,
            lineTransect
        };
        it("when using openStreetMap", () => __awaiter(void 0, void 0, void 0, function* () {
            const map = yield test_utils_1.createMap(Object.assign({ tileLayerName: "openStreetMap" }, options));
            const bounds = yield map.e("map.getBounds()");
            yield expect(bounds._northEast.lat).toBeGreaterThan(minNorth - 0.01); // 0.01 for Firefox inaccuracy
            yield expect(bounds._southWest.lat).toBeLessThan(minNorth);
            yield expect(bounds._northEast.lat).toBeGreaterThan(north - 0.01); // 0.01 for Firefox inaccuracy
            yield expect(bounds._southWest.lat).toBeLessThan(north);
            yield expect(bounds._northEast.lng).toBeGreaterThan(minEast);
            yield expect(bounds._southWest.lng).toBeLessThan(minEast);
            yield expect(bounds._northEast.lng).toBeGreaterThan(east);
            yield expect(bounds._southWest.lng).toBeLessThan(east);
        }));
        it("when using taustakartta", () => __awaiter(void 0, void 0, void 0, function* () {
            const map = yield test_utils_1.createMap(Object.assign({ tileLayerName: "taustakartta" }, options));
            const bounds = yield map.e("map.getBounds()");
            yield expect(bounds._northEast.lat).toBeGreaterThan(minNorth - 0.01); // 0.01 for Firefox inaccuracy
            yield expect(bounds._southWest.lat).toBeLessThan(minNorth);
            yield expect(bounds._northEast.lat).toBeGreaterThan(north - 0.01); // 0.01 for Firefox inaccuracy
            yield expect(bounds._southWest.lat).toBeLessThan(north);
            yield expect(bounds._northEast.lng).toBeGreaterThan(minEast);
            yield expect(bounds._southWest.lng).toBeLessThan(minEast);
            yield expect(bounds._northEast.lng).toBeGreaterThan(east);
            yield expect(bounds._southWest.lng).toBeLessThan(east);
        }));
    });
});
it("Falls back to center when no data and zoomToData given", () => __awaiter(void 0, void 0, void 0, function* () {
    const latLng = { lat: 25, lng: 60 };
    const options = {
        center: latLng,
        zoomToData: true,
        data: { type: "FeatureCollection", features: [] }
    };
    const map = yield test_utils_1.createMap(Object.assign({}, options));
    const center = yield map.e("map.getCenter()");
    yield expect(center.lat).toBe(latLng.lat);
    yield expect(center.lng).toBe(latLng.lng);
}));
it("Falls back to center when no zoomToData given", () => __awaiter(void 0, void 0, void 0, function* () {
    const latLng = { lat: 25, lng: 60 };
    const options = {
        center: latLng,
        data: { geoData: { type: "Point", coordinates: [45, 70] } }
    };
    const map = yield test_utils_1.createMap(Object.assign({}, options));
    const center = yield map.e("map.getCenter()");
    yield expect(center.lat).toBe(latLng.lat);
    yield expect(center.lng).toBe(latLng.lng);
}));
it("keeps finnish tileLayer if center is outside Finland but zoomToData causes view to initialize into Finland", () => __awaiter(void 0, void 0, void 0, function* () {
    const congo = {
        lat: 79.3499057749654,
        lng: 21.160612106323246
    };
    const data = { geoData: { type: "Point", coordinates: [25, 60] } };
    const map = yield test_utils_1.createMap({ tileLayerName: "taustakartta", zoom: 4, center: congo, data, zoomToData: true });
    yield expect(yield map.e("tileLayerName")).toBe("taustakartta");
}));
